#include "Bdef.h"

#if (INTFACE == C_CALL)
void Cigebr2d(int ConTxt, char *scope, char *top, int m, int n, int *A,
              int lda, int rsrc, int csrc)
#else
void igebr2d_(int *ConTxt, char *scope, char *top, int *m, int *n, int *A,
              int *lda, int *rsrc, int *csrc)
#endif
/*
 *  -- V1.1 BLACS routine --
 *  University of Tennessee, May 1, 1996
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Broadcast/receive for general integer arrays.
 *
 *  Arguments
 *  =========
 *
 *  ConTxt  (input) Ptr to int
 *          Index into MyConTxts00 (my contexts array).
 *
 *  SCOPE   (input) Ptr to char
 *          Limit the scope of the operation.
 *          = 'R' :   Operation is performed by a process row.
 *          = 'C' :   Operation is performed by a process column.
 *          = 'A' :   Operation is performed by all processes in grid.
 *
 *  TOP     (input) Ptr to char
 *          Controls fashion in which messages flow within the operation.
 *
 *  M       (input) Ptr to int
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) Ptr to int
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  A       (output) Ptr to integer two dimensional array
 *          The m by n matrix A.  Fortran77 (column-major) storage
 *          assumed.
 *
 *  LDA     (input) Ptr to int
 *          The leading dimension of the array A.  LDA >= M.
 *
 *
 *  RSRC    (input) Ptr to int
 *          The process row of the source of the matrix.
 *
 *  CSRC    (input) Ptr to int
 *          The process column of the source of the matrix.
 *
 *
 * ------------------------------------------------------------------------
 */
{
#if (BlacsDebugLvl > 0)
   void ArgCheck00(int, int, char *, char, char, char, int, int, int, int,
                   int *, int *);
#endif
   void mvcopy4(int, int, int *, int, int *);
   void vmcopy4(int, int, int *, int, int *);
   int Ahyp_br(BLACSCONTEXT *, char, BLACBUFF *, int, int, int);
   int Aidring_br(BLACSCONTEXT *, char, BLACBUFF *, int, int, int, int);
   int Asring_br(BLACSCONTEXT *, char, BLACBUFF *, int, int, int);
   int Ampath_br(BLACSCONTEXT *, char, BLACBUFF *, int, int, int, int);
   int Atree_br(BLACSCONTEXT *, char, BLACBUFF *, int, int, int, int);
   int Shyp_br(BLACSCONTEXT *, char, char *, int, int, int);
   int Sidring_br(BLACSCONTEXT *, char, char *, int, int, int, int);
   int Ssring_br(BLACSCONTEXT *, char, char *, int, int, int);
   int Smpath_br(BLACSCONTEXT *, char, char *, int, int, int, int);
   int Stree_br(BLACSCONTEXT *, char, char *, int, int, int, int);
   void UpdateBuffs(BLACBUFF *);
   BLACBUFF *getbuff(int);
   int BuffIsFree(BLACBUFF *, int);
   BLACSCONTEXT *ctxt;
   BLACBUFF *bp=NULL;
   int length;
   char ttop, tscope;
   int tlda, error;
   extern BLACBUFF *ActiveQ00;

   MGetConTxt(Mpval(ConTxt), ctxt);
   ttop = Mlowcase(*top);
   tscope = Mlowcase(*scope);
#if (BlacsDebugLvl > 0)
   ArgCheck00(Mpval(ConTxt), RT_BR, __FILE__, tscope, 'u', 'u', Mpval(m),
              Mpval(n), Mpval(lda), 1, Mpaddress(rsrc), Mpaddress(csrc));
#endif
   if (Mpval(m) <= Mpval(lda)) tlda = Mpval(lda);
   else tlda = Mpval(m);

   length = Mpval(m) * Mpval(n) * sizeof(int);
   if ( (Mpval(m) < tlda) && (Mpval(n) > 1) ) bp = getbuff(length);

   switch(ttop)
   {
   case ' ':
      if (bp == NULL)
         error = Stree_br(ctxt, tscope, (char *) A, length, 2,
                          Mpval(rsrc), Mpval(csrc));
      else
         error = Atree_br(ctxt, tscope, bp, length, 2,
                          Mpval(rsrc), Mpval(csrc));
      break;
   case 'h':
      if (bp == NULL)
      {
         error = Shyp_br(ctxt, tscope, (char *) A, length,
                         Mpval(rsrc), Mpval(csrc));
         if (error == NPOW2)
            error = Stree_br(ctxt, tscope, (char *) A, length, 2,
                             Mpval(rsrc), Mpval(csrc));
      }
      else
      {
         error = Ahyp_br(ctxt, tscope, bp, length,
                         Mpval(rsrc), Mpval(csrc));
         if (error == NPOW2)
            error = Atree_br(ctxt, tscope, bp, length, 2,
                             Mpval(rsrc), Mpval(csrc));
      }
      break;
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
      if (bp == NULL)
         error = Stree_br(ctxt, tscope, (char *) A, length, ttop-47,
                          Mpval(rsrc), Mpval(csrc));
      else
         error = Atree_br(ctxt, tscope, bp, length, ttop-47,
                          Mpval(rsrc), Mpval(csrc));
      break;
   case 't':
      if (bp == NULL)
         error = Stree_br(ctxt, tscope, (char *) A, length, ctxt->Nb_bs,
                          Mpval(rsrc), Mpval(csrc));
      else
         error = Atree_br(ctxt, tscope, bp, length, ctxt->Nb_bs,
                          Mpval(rsrc), Mpval(csrc));
      break;
   case 'i':
      if (bp == NULL)
         error = Sidring_br(ctxt, tscope, (char *) A, length, Mpval(rsrc),
                            Mpval(csrc), 1);
      else
         error = Aidring_br(ctxt, tscope, bp, length, Mpval(rsrc),
                            Mpval(csrc), 1);
      break;
   case 'd':
      if (bp == NULL)
         error = Sidring_br(ctxt, tscope, (char *) A, length, Mpval(rsrc),
                            Mpval(csrc), -1);
      else
         error = Aidring_br(ctxt, tscope, bp, length, Mpval(rsrc),
                            Mpval(csrc), -1);
      break;
   case 's':
      if (bp == NULL)
         error = Ssring_br(ctxt, tscope, (char *) A, length,
                           Mpval(rsrc), Mpval(csrc));
      else
         error = Asring_br(ctxt, tscope, bp, length,
                           Mpval(rsrc), Mpval(csrc));
      break;
   case 'm':
      if (bp == NULL)
         error = Smpath_br(ctxt, tscope, (char *) A, length, ctxt->Nr_bs,
                           Mpval(rsrc), Mpval(csrc));
      else
         error = Ampath_br(ctxt, tscope, bp, length, ctxt->Nr_bs,
                           Mpval(rsrc), Mpval(csrc));
      break;
   case 'f':
      if (bp == NULL)
         error = Smpath_br(ctxt, tscope, (char *) A, length, FULLCON,
                           Mpval(rsrc), Mpval(csrc));
      else
         error = Ampath_br(ctxt, tscope, bp, length, FULLCON,
                           Mpval(rsrc), Mpval(csrc));
      break;
   default :
      BlacsErr(Mpval(ConTxt), __LINE__, __FILE__, "Unknown topology '%c'",*top);
   }

#if (BlacsDebugLvl > 0)
   if (error == BADSCP)
      BlacsErr(Mpval(ConTxt), __LINE__, __FILE__, "Unknown scope '%c'", *scope);
#endif
/*
 * If we buffered, unpack.
 */
   if (bp)
   {
      ivmcopy(Mpval(m), Mpval(n), A, tlda, bp->Buff);
      UpdateBuffs(bp);
   }
   else if (ActiveQ00) UpdateBuffs(NULL);
}
